from typing import Optional

class TlfuCore:
    def __init__(self, size: int): ...
    def schedule(self, key: str, expire: int): ...
    def deschedule(self, key: str): ...
    def set_policy(self, key: str) -> Optional[str]: ...
    def set(self, key: str, expire: int) -> Optional[str]: ...
    def remove(self, key: str): ...
    def access(self, key: str): ...
    def advance(self, now: int, cache: dict): ...

class LruCore:
    def __init__(self, size: int): ...
    def schedule(self, key: str, expire: int): ...
    def deschedule(self, key: str): ...
    def set_policy(self, key: str) -> Optional[str]: ...
    def set(self, key: str, expire: int) -> Optional[str]: ...
    def remove(self, key: str): ...
    def access(self, key: str): ...
    def advance(self, now: int, cache: dict): ...

class BloomFilter:
    def put(self, key: str): ...
    def contains(self, key: str) -> bool: ...
